# This adds upfront requirements on vendored code found under lib/vendor/x
# Add one requirement per vendored package (or a comment if it is loaded on demand).

# The vendored library 'rgen' is loaded on demand.
