# Puppet as a Library "PAL"
# This is the entry point when using PAL as a standalone library.
#
# This requires all of puppet because 'settings' and many other things are still required in PAL.
# Eventually that will not be the case.
#
require_relative 'puppet'
require_relative 'puppet/pal/pal_api'
